/* ======================================================================
 *
 *  CCCCC          vr_input.joydev.c
 * CC   CC         Author(s): Bill Sherman
 * CC              Created: January 17, 2003
 * CC   CC         Last Modified: March 30, 2010
 *  CCCCC
 *
 * Code file for FreeVR inputs from the Linux Joystick input interface.
 *
 * Copyright 2010, Bill Sherman & Friends, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/*************************************************************************

USAGE:
	This device driver currently only works on Linux systems (because
	those are where the "joydev" interface exist -- though it may also
	be available on FreeBSD and similar systems).

	To use the "joydev" interface, the "joydev" module must be loaded.
	And, for 2.4 editions of the Linux kernel, if the device is connected
	to a USB port, then the "usb-ohci" module should be loaded (or perhaps
	the "usb-uhci" instead).  Game controllers connected to a combination
	Soundcard/Midi/Gameport card can also be used by installing the
	"gameport", "analog" and appropriate card module (in my case with
	an SB-Live card "emu1k1-gp").

	For 2.6 editions of the Linux kernel (or at least with the Fedora
	Core-2 distribution), there is a "ohci_hcd" module which seems to
	be loaded by default, and the "joydev" module should load when
	the device is plugged into the USB port.

	To test the device, the following command can be used (or similar
	if there is more than one USB human-inteface device connected):
		% jstest --event /dev/input/js0


	The configuration of the inputs is similar to all input devices:

	Inputs are specified with the "input" option:
		input "<name>" = "2switch(button[<number>])";
		input "<name>" = "valuator(axis[<number>])";
		input "<name>" = "6sensor(sim6[<number>])";

	Controls are specified in the freevrrc file:
		control "<control option>" = "2switch(button[<number>])";
		control "<control option>" = "valuator(axis[<number>])";

	Here are the available (2switch) control options for FreeVR:
	  :-(	"print_help" -- print info on how to use the input device
		"system_pause_toggle" -- toggle the system pause flag
		"print_context" -- print the overall FreeVR context data structure (for debugging)
		"print_config" -- print the overall FreeVR config data structure (for debugging)
		"print_input" -- print the overall FreeVR input data structure (for debugging)
		"print_struct" -- print the internal Joydev data structure (for debugging)

		"sensor_next" -- jump to the next 6-sensor on the list
		"setsensor(<num>)" -- set the simulated sensor to a particular one
		"sensor_reset" -- reset the current 6-sensor
		"sensor_resetall" -- reset the all the 6-sensors
		"temp_valuator" -- temporarily disable 6-sensor for valuator values
		"toggle_valuator" -- toggle valuator values instead of 6-sensor
		"temp_valuator_only" -- temporarily use translation values for valuator
		"toggle_valuator_only" -- toggle whether translation is saved for valuator
		"toggle_relative" -- toggle whether movement is relative to sensor's position
		"toggle_space_limit" -- toggle whether 6-sensor can go outside working volume
		"toggle_return_to_zero" -- toggle whether return-to-zero operation is on
		"toggle_swap_transrot" -- toggle whether to swap translation and rotation inputs
		"toggle_swap_yz" -- toggle the swapping of Y and Z translational movements

	Here are the available (valuator) control options for FreeVR:
		"set_transx" -- set the translation value for the simulated 6-sensor
		"set_transy" -- set the translation value for the simulated 6-sensor
		"set_transz" -- set the translation value for the simulated 6-sensor
		"set_azim" -- set the azimuth rotation value for the simulated 6-sensor
		"set_elev" -- set the elevation rotation value for the simulated 6-sensor
		"set_roll" -- set the roll rotation value for the simulated 6-sensor

	Here are the FreeVR configuration argument options for the Joydev:
		"devname" - /dev/input/js<num> port Joydev is connected to
			("/dev/input/js0" is the default)
		"restrict" - limit simulated 6-sensors to move within the working-volume
		"valuatoroverride" - produce valuator values instead of moving the 6-sensor
		"returntozero" - valuator offsets map to absolute or relative position of simulated 6-sensor
		"relativerot" - rotation of simulated 6-sensor is based on sensor's or world's coord-sys
	  	"swaptransrot" -- swap the translational and rotational inputs
		"workingvolume" - orthogonally defined volume specifying the valid range of simulated 6-sensor movement
		"transscale" - scaling factor used to tune the translational movement of a simulated 6-sensor
		"rotscale" - scaling factor used to tune the rotational movement of a simulated 6-sensor
		"valscale" - scaling factor used to tune the valuator input range

HISTORY:
	17 January 2003 (Bill Sherman) -- wrote initial version using some
		example code from the net and Albert.

	21 January 2003 (Bill Sherman) -- wrote interface to FreeVR.  Added
		compile time flags to make a version that will compile on
		non-Linux systems, but without providing any real input.

	22 January 2003 (Bill Sherman) -- wrote the simulated 6-sensor
		interface.

	21-23 April 2003 (Bill Sherman)
		Updated to new input matching function code style.  Changed "opaque"
		  field to "aux_data".  Split _JoydevFunction() into 5 functions.
		  Added new vrPrintStyle argument to _JoydevPrintStruct() for the
		  sake of the new "PrintAux" callback.

	1 May 2003 (Bill Sherman)
		A simulated 6-sensor input is now only assigned when the values
		  change.

	3 June 2003 (Bill Sherman)
		Now include "vr_enums.h" for the TEST_APP code.
		Added the address of the auxillary data to the printout.
		Added the "system_pause_toggle" control callback.

	9 October 2003 (Bill Sherman)
		Added inclusion of "sys/time.h" when compiling under Linux
		because *older* versions of Linux require it -- it seems
		to be included in some of the other header files on newer
		Linuxii.

       13 October 2003 (John Stone)
		Added inclusion of "sys/ioctl.h" when compiling under Linux
		because *older* versions of Linux require it -- it seems
		to be included in some of the other header files on newer
		Linuxii.

       16 November 2004 (Bill Sherman & Jeff Stuart)
		Fixed some null dereferencing bugs for which it isn't clear
		how it worked in the past.

       22 November 2004 (Bill Sherman & Dave Zielinski)
		I added some protection against getting event inputs outside
		the bounds allocated for buttons or axes.  I also added a
		fix sent to me by Dave on 9/22/2004 (had "num_axes", s/b
		"num_buttons").

	16 October 2009 (Bill Sherman)
		A quick fix to the _JoydevParseArgs() routine to handle the no-arguments
		case.

	30 March 2010 (Bill Sherman)
		A quick fix to prevent the "valuator_inputs" field from being
		written beyond its bounds.  Also changed references to "/dev/js0"
		to be "/dev/input/js0" (including the default port) since that
		seems to be how modern Linuxii have moved.

TODO:
	- Bring in some of the improved comments, etc. from VruiDD (and the new skeleton) [10/19/2009]

	- Should check to see whether some other free Unixii have the
		joydev interface.

	- Make sure error handling of controls & input type mismatches is
		reasonable -- eg. trying to use a valuator control with a
		button input & vice versa.

	- When a button goes to a valuator, allow sign change operation.
		(ie. allow the sign to be specified in the "control" config line.)

	- Allow multiple inputs to same control, so for example set_roll can
		be implemented by two opposing button presses.

	- make a daemon interface [03/30/2010: Why?  Or at least, why a special
		daemon interface just for joydev?]

	- make the rumble pak stuff work -- once the Linux kernel can handle it

	- (perhaps) make use of incoming time data

**************************************************************************/
#if !defined(FREEVR) && !defined(TEST_APP) && !defined(CAVE)
/* Choose how this code should be compiled (ie. for CAVE, standalone, etc). */
#  define	FREEVR
#  undef	TEST_APP
#  undef	CAVE
#endif

#undef	DEBUG


#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>		/* needed for open() and fcntl() */
#ifdef __linux
#  include <sys/ioctl.h>
#  include <sys/time.h>		/* needed for struct timeval */
#  include <linux/joystick.h>	/* needed for Joydev flags and masks */
#endif

#include "vr_debug.h"
#if 0 /* TODO: unsure about utils.h (perhaps delete) */
#include "vr_utils.h"
#endif

#if defined (FREEVR)
#  include "vr_input.h"
#  include "vr_input.opts.h"
#  include "vr_parse.h"
#  include "vr_shmem.h"
#endif


#if defined(TEST_APP) || defined(CAVE)
#  define	VR_X	0
#  define	VR_Y	1
#  define	VR_Z	2
#  define	VR_W	3
#  define	VR_AZIM	0
#  define	VR_ELEV	1
#  define	VR_ROLL	2
#endif

#if defined(TEST_APP)
#  include "vr_serial.c"
#  include "vr_enums.h"
#  define	vrShmemAlloc	malloc
#  define	vrShmemAlloc0	malloc
#  define	vrAtoI		atoi
#endif


#if defined(CAVE) /* { */
#  include <math.h>
#  include "cave.h"
#  include "cave.private.h"
#  include "cave.tracker.h"
static CAVE_SENSOR_ST initial_cave_sensor = {
		0,5,0,
		0,0,0,
		0,	/*timestamp*/
		FALSE,	/*calibrated*/
		CAVE_TRACKER_FRAME
	};
#endif /* } CAVE */


	/***************************************************/
	/*** definitions for interfacing with the device ***/
	/***                                             ***/
#define	DEFAULT_DEVICE	"/dev/input/js0"


/** Joydev indicies **/
/* NOTE: These labels are based on the Logitech Wingman USB controller, they may not reflect other controllers */
/*   Additional comments are to compare with other devices */
						/* 4-but analog		GrIP  */
#define Joydev_BUTTONINDEX_A	 0		/* red			green */
#define Joydev_BUTTONINDEX_B	 1		/* blue			yellow */
#define Joydev_BUTTONINDEX_C	 2		/* yellow		red */
#define Joydev_BUTTONINDEX_X	 3		/* green		blue */
#define Joydev_BUTTONINDEX_Y	 4		/*			L1 */
#define Joydev_BUTTONINDEX_Z	 5		/*			R1 */
#define Joydev_BUTTONINDEX_L1	 6		/*			L2 */
#define Joydev_BUTTONINDEX_R1	 7		/*			R2 */
#define Joydev_BUTTONINDEX_START 8		/*			Start */
#define Joydev_BUTTONINDEX_L2	 9		/*			Select */
#define Joydev_BUTTONINDEX_R2	10
#define Joydev_BUTTONINDEX_MODE	11

#define	Joydev_VALUATORINDEX_LEFT_X	0
#define	Joydev_VALUATORINDEX_LEFT_Y	1
#define	Joydev_VALUATORINDEX_SLIDER	2
#define	Joydev_VALUATORINDEX_RIGHT_X	3
#define	Joydev_VALUATORINDEX_RIGHT_Y	4
#define	Joydev_VALUATORINDEX_DIGITAL_X	5	/* joystick-X		joystick-X */
#define	Joydev_VALUATORINDEX_DIGITAL_Y	6	/* joystick-Y		joystick-Y */

/* Joydev sensitivity values */
#define TRANS_SENSITIVITY	0.1
#define ROT_SENSITIVITY		1.0
#define VALUATOR_SENSITIVITY	1.0


/****************************************************************/
/*** auxillary structure of the current data from the device. ***/
typedef struct {
		/* these are for interfacing with the hardware */
		int		fd;		/* communication file descriptor */
		fd_set		read_fds;	/* list of file descriptors for select() -- will only use one */
		char		*devfile;	/* name of device file */
		int		open;		/* flag with Joydev successfully open */
		char		name[256];	/* name reported by the device */

		/* these are for internal data parsing */
		char		version[256];	/* self-reported version of the device */

#ifdef CAVE
		/* CAVE specific fields here */

#elif defined(FREEVR)
		/* FREEVR specific fields here */

#define MAX_BUTTONS    32
		vr2switch	*button_inputs[MAX_BUTTONS];
#define MAX_VALUATORS  16
		vrValuator	*valuator_inputs[MAX_VALUATORS];
		float		valuator_sign[MAX_VALUATORS];
#define MAX_6SENSORS  16
		vr6sensor	*sensor6_inputs[MAX_6SENSORS];
		int		sim6sensor_change;	/* flag that indicates whether any of the controls to a simulated 6-sensor have changed */

		/* for the FreeVR simulated 6-sensors */
		int		active_sensor;		/* The simulated 6-sensor that is being actively controlled */
		vr6sensorConv	sensor6_options;	/* Structure of settings that affect how a 6-sensor is simulated */
		float		sim_values[6];		/* Array of 6 valuator values used to move the simulated 6-sensor */
#endif /* end library-specific fields */

		/* information about the current values */
		int		num_axes;	/* number of valuators (aka axes) reported by the device */
		float		*axis;		/* array of axis values */
		int		*axis_change;	/* array of flags to indicate new valuator value */
		int		num_buttons;	/* number of buttons reported by the device */
		int		*button;	/* array of button values */
		int		*button_change;	/* array of flags to indicate new button value */

		/* information about how to filter the data for use with the VR library (FreeVR or CAVE) */
		float		scale_valuator;	/* scaling factor for valuators */
		/* TODO: determine whether any filter values/settings are necessary */

	} _JoydevPrivateInfo;



	/*********************************************/
	/*** General NON public interface routines ***/
	/*********************************************/

/******************************************************/
/* typename is used to specify a particular device among many that */
/*   share (more or less) the same protocol.  The typename is then */
/*   used to determine what specific features are available with   */
/*   this particular type of device.                               */
static void _JoydevInitializeStruct(_JoydevPrivateInfo *aux, char *typename)
{
	aux->version[0] = '\0';
	aux->name[0] = '\0';

	aux->scale_valuator = VALUATOR_SENSITIVITY;	/* set the default valuator scaling factor */

#ifndef TEST_APP /* {  (we have no simulated inputs for the test app) */
	/* set values for simulated 6-sensor */
	aux->sim_values[0] = 0.0;			/* start with no movement values in the valuators holders */
	aux->sim_values[1] = 0.0;
	aux->sim_values[2] = 0.0;
	aux->sim_values[3] = 0.0;
	aux->sim_values[4] = 0.0;
	aux->sim_values[5] = 0.0;

	aux->active_sensor = 1;				/* start with the "wand" active  -- potential bug */
	aux->active_sensor = 0;				/* TODO: go back to wand as default -- currently head (ie. delete this line) */

	aux->sensor6_options.azimuth_axis = VR_Y;	/* azimuth is about the Y axis */
	aux->sensor6_options.relative_axis = 1;		/* default to relative rotations */
	aux->sensor6_options.return_to_zero = 0;	/* default to free floating */
	aux->sensor6_options.ignore_all = 0;		/* default to joystick off */
	aux->sensor6_options.ignore_trans = 0;		/* default to joystick off */
	aux->sensor6_options.restrict_space = 1;	/* default restricted to working-volume */
	aux->sensor6_options.swap_yz = 0;		/* dafault to standard Y & Z axes */
	aux->sensor6_options.swap_transrot = 0;		/* default to unswapped translation & rotation */
	aux->sensor6_options.trans_scale = TRANS_SENSITIVITY;	/* set the default translational scaling */
	aux->sensor6_options.rot_scale = ROT_SENSITIVITY;	/* set the default rotational scaling */

	/* The default working-volume is that of the typical CAVE */
	aux->sensor6_options.working_volume_min[VR_X] = -5.0;
	aux->sensor6_options.working_volume_max[VR_X] =  5.0;
	aux->sensor6_options.working_volume_min[VR_Y] =  0.0;
	aux->sensor6_options.working_volume_max[VR_Y] = 10.0;
	aux->sensor6_options.working_volume_min[VR_Z] = -5.0;
	aux->sensor6_options.working_volume_max[VR_Z] =  5.0;
#endif /* } */

	/* everything else is zero'd by default */
}


/******************************************************/
static void _JoydevPrintStruct(FILE *file, _JoydevPrivateInfo *aux, vrPrintStyle style)
{
	int	count;

	vrFprintf(file, "Joydev device internal structure (0x%p):\n", aux);
	vrFprintf(file, "\r\tversion -- '%s'\n", aux->version);
	vrFprintf(file, "\r\tname -- '%s'\n", aux->name);
	vrFprintf(file, "\r\tfd = %d\n\tdevfile = '%s'\n\topen = %d\n",
		aux->fd,
		aux->devfile,
		aux->open);

	/* print the raw values */
	vrFprintf(file, "\r\tnum_buttons = %d\n", aux->num_buttons);
	for (count = 0; count < aux->num_buttons; count++)
		vrFprintf(file, "\r\t\tbutton[%2d] = %d  (changed? %d)\n", count, aux->button[count], aux->button_change[count]);

	vrFprintf(file, "\r\tnum_valuators = %d\n", aux->num_axes);
	for (count = 0; count < aux->num_axes; count++)
		vrFprintf(file, "\r\t\tvaluator[%d] = %+f  (changed? %d)\n", count, aux->axis[count], aux->axis_change[count]);

#ifdef FREEVR /* { */
	/* print the FreeVR input objects */
	vrFprintf(file, "\r\tbutton inputs:\n");
	for (count = 0; count < MAX_BUTTONS; count++) {
		vrFprintf(file, "\r\t\tbutton_input[%d] = 0x%p", count, aux->button_inputs[count]);
		if (aux->button_inputs[count] != NULL)
			vrFprintf(file, " type: %s, name: %s\n", vrInputTypeName(aux->button_inputs[count]->input_type), aux->button_inputs[count]->my_object->name);
		else	vrFprintf(file, "\n");
	}

	vrFprintf(file, "\r\tvaluator inputs:\n");
	for (count = 0; count < MAX_VALUATORS; count++) {
		vrFprintf(file, "\r\t\tvaluator_input[%d] = 0x%p", count, aux->valuator_inputs[count]);
		if (aux->valuator_inputs[count] != NULL)
			vrFprintf(file, " type: %s, name: %s\n", vrInputTypeName(aux->valuator_inputs[count]->input_type), aux->valuator_inputs[count]->my_object->name);
		else	vrFprintf(file, "\n");
	}

	vrFprintf(file, "\r\tsensor6 inputs (active = %d):\n", aux->active_sensor);
	for (count = 0; count < MAX_6SENSORS; count++) {
		vrFprintf(file, "\r\t\tsensor6_inputs[%d] = 0x%p", count, aux->sensor6_inputs[count]);
		if (aux->sensor6_inputs[count] != NULL)
			vrFprintf(file, " type: %s, name: %s\n", vrInputTypeName(aux->sensor6_inputs[count]->input_type), aux->sensor6_inputs[count]->my_object->name);
		else	vrFprintf(file, "\n");
	}

	vrFprintf(file, "\r"
			"\tsensor6_options = %d, %d, %d, %d, %d, %d, %d, %d, %d, %.2f %.2f (%.2f--%.2f %.2f--%.2f %.2f--%.2f)\n",
			aux->sensor6_options.azimuth_axis,
			aux->sensor6_options.ignore_all,
			aux->sensor6_options.ignore_trans,
			aux->sensor6_options.tmp_ignore_trans,
			aux->sensor6_options.relative_axis,
			aux->sensor6_options.return_to_zero,
			aux->sensor6_options.restrict_space,
			aux->sensor6_options.swap_yz,
			aux->sensor6_options.swap_transrot,
			aux->sensor6_options.trans_scale,
			aux->sensor6_options.rot_scale,
			aux->sensor6_options.working_volume_min[0],
			aux->sensor6_options.working_volume_max[0],
			aux->sensor6_options.working_volume_min[1],
			aux->sensor6_options.working_volume_max[1],
			aux->sensor6_options.working_volume_min[2],
			aux->sensor6_options.working_volume_max[2]);
	vrFprintf(file, "\r"
			"\tsim_values = (%.2f, %.2f, %.2f,  %.2f, %.2f, %.2f)\n"
			"\tsim6sensor_change = %d\n",
			aux->sim_values[VR_X],
			aux->sim_values[VR_Y],
			aux->sim_values[VR_Z],
			aux->sim_values[VR_AZIM+3],
			aux->sim_values[VR_ELEV+3],
			aux->sim_values[VR_ROLL+3],
			aux->sim6sensor_change);
#endif /* } FREEVR */

}


/**************************************************************************/
static void _JoydevPrintHelp(FILE *file, _JoydevPrivateInfo *aux)
{
	vrFprintf(file, BOLD_TEXT "Sorry, Joydev - print_help control not yet implemented.\n" NORM_TEXT);
}


/******************************************************/
/* Returns the number of inputs sucessfully read. */
static int _JoydevReadInput(_JoydevPrivateInfo *aux)
{
	char			inputs_read = 0;
#ifdef __linux /* { */
static	struct timeval		no_block = { 0 , 0 };
static	struct js_event		event;

	if (aux->open == 0) {
		return (0);
	}

	/* set the file descriptor list to watch the joydev device */
	FD_ZERO(&aux->read_fds);
	FD_SET(aux->fd, &aux->read_fds);

	/* loop until there is no more data on the device */
	while (select(aux->fd + 1, &aux->read_fds, NULL, NULL, &no_block)) {

		/* read the data from the device */
		read(aux->fd, &event, sizeof(event));

#if 0
		vrPrintf("Event: time = %d, type = %d, code = %d, value = %d\n", event.time, event.type, event.number, event.value);
#endif

		/* parse the event */
		if (event.type & JS_EVENT_INIT) {
			/* got an initial value */
#if 0
			vrPrintf("Yo: initial value recieved\n");
#endif
		}
		switch (event.type & ~JS_EVENT_INIT) {
		case JS_EVENT_AXIS:
			if (event.number+1 > aux->num_axes) {
				vrDbgPrintfN(AALWAYS_DBGLVL, RED_TEXT "Unexpected axis %d (only %d available) -- ignoring input.\n" NORM_TEXT, event.number, aux->num_axes);
			} else {
				vrDbgPrintfN(INPUT_DBGLVL, "_JoydevReadInput(): got axis event: number = %d, value = %d\n", event.number, event.value);
				aux->axis[event.number] = event.value / 32767.0;
				aux->axis_change[event.number] = 1;
				inputs_read++;
			}
			break;
		case JS_EVENT_BUTTON:
			if (event.number+1 > aux->num_buttons) {
				vrDbgPrintfN(AALWAYS_DBGLVL, RED_TEXT "Unexpected button %d (only %d available) -- ignoring input.\n" NORM_TEXT, event.number, aux->num_buttons);
			} else {
				vrDbgPrintfN(INPUT_DBGLVL, "_JoydevReadInput(): got button event: number = %d, value = %d\n", event.number, event.value);
				aux->button[event.number] = event.value;
				aux->button_change[event.number] = 1;
				inputs_read++;
			}
			break;
		}
	}
#endif /* } */

	return (inputs_read);
}


/**********************************************************/
/* _JoydevInitializeDevice() is called in the OPEN phase  */
/*   of input interface -- after the types of inputs have */
/*   been determined (during the CREATE phase).           */
static int _JoydevInitializeDevice(_JoydevPrivateInfo *aux)
{
#ifdef __linux /* { */
	unsigned long	js_version;
#endif /* } */

	if (aux == NULL)
		return -1;

#ifdef __linux /* { */
	/* read name and parameters of device */
	ioctl(aux->fd, JSIOCGVERSION, &js_version);
	ioctl(aux->fd, JSIOCGNAME(255), &aux->name);
	ioctl(aux->fd, JSIOCGAXES, &aux->num_axes);
	ioctl(aux->fd, JSIOCGBUTTONS, &aux->num_buttons);

	sprintf(aux->version, "Linux Joystick Driver version 0x%lx", js_version);
	vrDbgPrintfN(INPUT_DBGLVL, "Initializing for Joystick Driver version 0x%x, name = '%s', with %d axes and %d buttons, aux->fd = %d\n", js_version, aux->name, aux->num_axes, aux->num_buttons, aux->fd);
#else /* } { */
	sprintf(aux->version, "Non-Linux system: No Linux Joystick Driver version available");
	sprintf(aux->name, "Non-Linux system: Unable to obtain device name");
	aux->num_axes = 0;
	aux->num_buttons = 0;
#endif /* } */

	/* allocate memory for reading values from the device */
	aux->axis = vrShmemAlloc0(aux->num_axes * sizeof(float));
	aux->axis_change = vrShmemAlloc0(aux->num_axes * sizeof(int));
	aux->button = vrShmemAlloc0(aux->num_buttons * sizeof(int));
	aux->button_change = vrShmemAlloc0(aux->num_buttons * sizeof(int));

	/* read some initial data */
	_JoydevReadInput(aux);

	return 0;
}


/******************************************************/
/* Translate a string name of a button (the "instance" config) into a numeric value */
static unsigned int _JoydevButtonValue(char *buttonname)
{
	switch (buttonname[0]) {
	case '0':		/* for buttons that start with a digit, assume it's just a numeric reference */
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		return vrAtoI(buttonname);

	case 'a':
	case 'A':
		return Joydev_BUTTONINDEX_A;

	case 'b':
	case 'B':
		return Joydev_BUTTONINDEX_B;

	case 'c':
	case 'C':
		return Joydev_BUTTONINDEX_C;

	case 'x':
	case 'X':
		return Joydev_BUTTONINDEX_X;

	case 'y':
	case 'Y':
		return Joydev_BUTTONINDEX_Y;

	case 'z':
	case 'Z':
		return Joydev_BUTTONINDEX_Z;

	case 'l':
	case 'L':
		switch (buttonname[1]) {
		case '1':
			return Joydev_BUTTONINDEX_L1;
		case '2':
			return Joydev_BUTTONINDEX_L2;
		default:
			return -1;
		}

	case 's':	/* select and start */
	case 'r':	/* 1 and 2 */
	case 'R':
		switch (buttonname[1]) {
		case '1':
			return Joydev_BUTTONINDEX_R1;
		case '2':
			return Joydev_BUTTONINDEX_R2;
		default:
			return -1;
		}

	case 'S':
		return Joydev_BUTTONINDEX_START;

	case 'm':
	case 'M':
		return Joydev_BUTTONINDEX_MODE;
	}

	return -1;
}


/******************************************************/
/* Translate a string name of a valuator (the "instance" config) into a numeric value */
static unsigned int _JoydevValuatorValue(char *valuatorname)
{
	/* skip an opening sign character (is handled in _JoydevValuatorInput) */
	if (valuatorname[0] == '-' || valuatorname[0] == '+')
		valuatorname++;

	switch (valuatorname[0]) {
	case '0':		/* for valuators that start with a digit, assume it's just a numeric reference */
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		return vrAtoI(valuatorname);

	case 's':	/* slider */
	case 'S':
		return Joydev_VALUATORINDEX_SLIDER;

	case 'l':	/* Left Joystick, X and Y */
	case 'L':
		switch (valuatorname[strlen(valuatorname)-1]) {
		case 'x':
		case 'X':
			return Joydev_VALUATORINDEX_LEFT_X;
		case 'y':
		case 'Y':
			return Joydev_VALUATORINDEX_LEFT_Y;
		default:
			return -1;
		}

	case 'r':	/* Right Joystick, X and Y */
	case 'R':
		switch (valuatorname[strlen(valuatorname)-1]) {
		case 'x':
		case 'X':
			return Joydev_VALUATORINDEX_RIGHT_X;
		case 'y':
		case 'Y':
			return Joydev_VALUATORINDEX_RIGHT_Y;
		default:
			return -1;
		}

	case 'd':	/* Digital Joystick, X and Y */
	case 'D':
		switch (valuatorname[strlen(valuatorname)-1]) {
		case 'x':
		case 'X':
			return Joydev_VALUATORINDEX_DIGITAL_X;
		case 'y':
		case 'Y':
			return Joydev_VALUATORINDEX_DIGITAL_Y;
		default:
			return -1;
		}
	}

	return -1;
}


	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/


#if defined(FREEVR) /* { */
/*****************************************************************/
/*** Functions for FreeVR access of Joydev devices for input ***/
/*****************************************************************/


	/************************************/
	/***  FreeVR NON public routines  ***/
	/************************************/


/*********************************************************/
static void _JoydevParseArgs(_JoydevPrivateInfo *aux, char *args)
{
	float	volume_values[6];	/* for reading the working volume array */

	/* In the rare case of no arguments, just return */
	if (args == NULL)
		return;

	/*****************************************/
	/** Argument format: "devname" "=" file **/
	/**    or "devfile" "=" file            **/
	/*****************************************/
	vrArgParseString(args, "devname", &(aux->devfile));
	vrArgParseString(args, "devfile", &(aux->devfile));

	/* arguments for simulated 6-sensors */

	/***************************************************************/
	/** Argument format: "restrict" "=" { "on" | "off" | number } **/
	/***************************************************************/
	vrArgParseBool(args, "restrict", &(aux->sensor6_options.restrict_space));

	/***********************************************************************/
	/** Argument format: "valuatorOverride" "=" { "on" | "off" | number } **/
	/***********************************************************************/
	vrArgParseBool(args, "valuatoroverride", &(aux->sensor6_options.ignore_trans));

	/*******************************************************************/
	/** Argument format: "returnToZero" "=" { "on" | "off" | number } **/
	/*******************************************************************/
	vrArgParseBool(args, "returntozero", &(aux->sensor6_options.return_to_zero));

	/******************************************************************/
	/** Argument format: "relativeRot" "=" { "on" | "off" | number } **/
	/******************************************************************/
	vrArgParseBool(args, "relativerot", &(aux->sensor6_options.relative_axis));

	/*******************************************************************/
	/** Argument format: "swapTransRot" "=" { "on" | "off" | number } **/
	/*******************************************************************/
	vrArgParseBool(args, "swaptransrot", &(aux->sensor6_options.swap_transrot));

	/*******************************************************************/
	/** Argument format: "swapTransRot" "=" { "on" | "off" | number } **/
	/*******************************************************************/
	vrArgParseBool(args, "swaptransrot", &(aux->sensor6_options.swap_transrot));

	/******************************************************************************************/
	/** Argument format: "workingVolume" "=" <float> <float> <float> <float> <float> <float> **/
	/******************************************************************************************/
	if (vrArgParseFloatList(args, "workingvolume", volume_values, 6)) {
		aux->sensor6_options.working_volume_min[VR_X] = volume_values[0];
		aux->sensor6_options.working_volume_max[VR_X] = volume_values[1];
		aux->sensor6_options.working_volume_min[VR_Y] = volume_values[2];
		aux->sensor6_options.working_volume_max[VR_Y] = volume_values[3];
		aux->sensor6_options.working_volume_min[VR_Z] = volume_values[4];
		aux->sensor6_options.working_volume_max[VR_Z] = volume_values[5];
	}

	/**********************************************/
	/** Argument format: "transScale" "=" number **/
	/**********************************************/
	if (vrArgParseFloat(args, "transscale", &(aux->sensor6_options.trans_scale))) {
		aux->sensor6_options.trans_scale *= TRANS_SENSITIVITY;
	}

	/********************************************/
	/** Argument format: "rotScale" "=" number **/
	/********************************************/
	if (vrArgParseFloat(args, "rotscale", &(aux->sensor6_options.rot_scale))) {
		aux->sensor6_options.rot_scale *= ROT_SENSITIVITY;
	}

	/********************************************/
	/** Argument format: "valScale" "=" number **/
	/********************************************/
	if (vrArgParseFloat(args, "valscale", &(aux->scale_valuator))) {
		aux->scale_valuator *= VALUATOR_SENSITIVITY;
	}

	/** TODO: other arguments to parse go here **/
}


/************************************************************/
static void _JoydevGetData(vrInputDevice *devinfo)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;
	int			num_inputs;
	int			count;
	float			valuator_value;

	num_inputs = _JoydevReadInput(aux);

	if (num_inputs > 0) {

		/* handle button inputs as buttons or self-controls */
		for (count = 0; count < aux->num_buttons; count++) {
			if ((aux->button_change[count] != 0)) {
				if (aux->button_inputs[count] != NULL) {
					switch (aux->button_inputs[count]->input_type) {
					case VRINPUT_BINARY:
						vrAssign2switchValue((vr2switch *)(aux->button_inputs[count]), ((aux->button[count]) != 0));
						break;
					case VRINPUT_CONTROL:
						vrCallbackInvokeDynamic(((vrControl *)(aux->button_inputs[count]))->callback, 1, ((aux->button[count]) != 0));
						break;
					default:
						vrErrPrintf(RED_TEXT "_JoydevGetData: Unable to handle button inputs that aren't Binary or Control inputs\n" NORM_TEXT);
						break;
					}

				}
				aux->button_change[count] = 0;
			}
		}

		/* handle valuator inputs as valuators or self-controls */
		for (count = 0; count < aux->num_axes; count++) {
			if ((aux->axis_change[count] != 0)) {
				if (aux->valuator_inputs[count] != NULL) {
					valuator_value = aux->axis[count] * aux->scale_valuator * aux->valuator_sign[count];
					switch (aux->valuator_inputs[count]->input_type) {
					case VRINPUT_VALUATOR:
						vrAssignValuatorValue((vrValuator *)(aux->valuator_inputs[count]), valuator_value);
						break;
					case VRINPUT_CONTROL:
						vrCallbackInvokeDynamic(((vrControl *)(aux->valuator_inputs[count]))->callback, 1, &valuator_value);
						break;
					default:
						vrErrPrintf(RED_TEXT "_JoydevGetData: Unable to handle valuator inputs that aren't Floating or Control inputs\n" NORM_TEXT);
						break;
					}
				}
				aux->axis_change[count] = 0;
			}
		}
	}

	/* handle valuator inputs to be converted to simulated 6-sensor inputs */
	if (!aux->sensor6_options.ignore_all && devinfo->num_6sensors > 0 && aux->sim6sensor_change) {
		vrAssign6sensorValueFromValuators(&(devinfo->sensor6[aux->active_sensor]), aux->sim_values, &(aux->sensor6_options), -1 /* (ie. no change to oob flag) */);
		if ((aux->sim_values[VR_X] == 0)
		   && (aux->sim_values[VR_Y] == 0)
		   && (aux->sim_values[VR_Z] == 0)
		   && (aux->sim_values[VR_AZIM+3] == 0)
		   && (aux->sim_values[VR_ELEV+3] == 0)
		   && (aux->sim_values[VR_ROLL+3] == 0)) {
			aux->sim6sensor_change = 0;
		}
	}
}


	/****************************************************************/
	/*    Function(s) for parsing Joydev "input" declarations.      */
	/*                                                              */
	/*  These _Joydev<type>Input() functions are called during the  */
	/*  CREATE phase of the input interface.                        */

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

/**************************************************************************/
static vrInputMatch _JoydevButtonInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
        _JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;
	int			button_num;

	button_num = _JoydevButtonValue(dti->instance);
	if (button_num == -1) {
		vrErrPrintf("_JoydevButtonInput(): Warning, button['%s'] did not match any known button\n", dti->instance);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (aux->button_inputs[button_num] != NULL)
		vrErrPrintf(RED_TEXT "_JoydevButtonInput(): Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	aux->button_inputs[button_num] = (vr2switch *)input;
	vrDbgPrintfN(INPUT_DBGLVL, "_JoydevButtonInput(): assigned button event of value 0x%02x to input pointer = 0x%p)\n",
		button_num, aux->button_inputs[button_num]);

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}


/**************************************************************************/
static vrInputMatch _JoydevValuatorInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
        _JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;
	int			valuator_num;

	valuator_num = _JoydevValuatorValue(dti->instance);
	if (dti->instance[0] == '-')
		aux->valuator_sign[valuator_num] = -1.0;
	else	aux->valuator_sign[valuator_num] =  1.0;

	if (valuator_num == -1) {
		vrErrPrintf("_JoydevValuatorInput(): Warning, valuator['%s'] did not match any known valuator\n", dti->instance);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (aux->valuator_inputs[valuator_num] != NULL) {
		vrErrPrintf(RED_TEXT "_JoydevValuatorInput(): Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	}

	aux->valuator_inputs[valuator_num] = (vrValuator *)input;
	vrDbgPrintfN(INPUT_DBGLVL, "_JoydevValuatorInput(): assigned valuator event of value 0x%02x to input pointer = 0x%p)\n",
		valuator_num, aux->valuator_inputs[valuator_num]);

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}


/**************************************************************************/
static vrInputMatch _Joydev6sensorInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
        _JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;
	int			sensor_num;

	sensor_num = vrAtoI(dti->instance);
	if (sensor_num < 0) {
		vrErrPrintf("_Joydev6sensorInput(): Warning, sensor number must be between %d and %d\n", 0, MAX_6SENSORS);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (aux->sensor6_inputs[sensor_num] != NULL)
		vrErrPrintf(RED_TEXT "_Joydev6sensorInput(): Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	aux->sensor6_inputs[sensor_num] = (vr6sensor *)input;

	/* the vrAssign6sensorR2Exform() function does the parsing of the next token */
	vrAssign6sensorR2Exform(aux->sensor6_inputs[sensor_num], strchr(dti->instance, ','));

	vrDbgPrintfN(INPUT_DBGLVL, "_Joydev6sensorInput(): assigned 6sensor event of value 0x%02x to input pointer = 0x%p)\n",
		sensor_num, aux->sensor6_inputs[sensor_num]);

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}



	/************************************************************/
	/***************** FreeVR Callback routines *****************/
	/************************************************************/

	/************************************************************/
	/*    Callbacks for controlling the device's features.      */
	/*                                                          */

/* TODO: describe the consistent arguments of the callbacks. */
/*   ... devinfo -- same as "input", but also sometimes used to get */
/*     information about the device so it can be altered -- eg. _SpacetecSensorResetCallback()  and _XwindowsSensorNextCallback() */
/*       it is also frequently passed on to other callback routines which in */
/*       turn generally use it to access the "aux" data.  */

/************************************************************/
static void _JoydevSystemPauseToggleCallback(vrInputDevice *devinfo, int value)
{
        if (value == 0)
                return;

	devinfo->context->paused ^= 1;
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: system_pause = %d.\n",
		devinfo->context->paused);
}

/************************************************************/
static void _JoydevPrintContextStructCallback(vrInputDevice *devinfo, int value)
{
        if (value == 0)
                return;

        vrFprintContext(stdout, devinfo->context, verbose);
}

/************************************************************/
static void _JoydevPrintConfigStructCallback(vrInputDevice *devinfo, int value)
{
        if (value == 0)
                return;

        vrFprintConfig(stdout, devinfo->context->config, verbose);
}

/************************************************************/
static void _JoydevPrintInputStructCallback(vrInputDevice *devinfo, int value)
{
        if (value == 0)
                return;

        vrFprintInput(stdout, devinfo->context->input, verbose);
}

/************************************************************/
static void _JoydevPrintStructCallback(vrInputDevice *devinfo, int value)
{
        _JoydevPrivateInfo  *aux = (_JoydevPrivateInfo *)devinfo->aux_data;

        if (value == 0)
                return;

        _JoydevPrintStruct(stdout, aux, verbose);
}

/************************************************************/
static void _JoydevPrintHelpCallback(vrInputDevice *devinfo, int value)
{
        _JoydevPrivateInfo  *aux = (_JoydevPrivateInfo *)devinfo->aux_data;

        if (value == 0)
                return;

        _JoydevPrintHelp(stdout, aux);
}

/************************************************************/
static void _JoydevSensorNextCallback(vrInputDevice *devinfo, int value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	if (devinfo->num_6sensors == 0) {
		vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: next sensor -- no sensors available.\n",
			aux->active_sensor);
		return;
	}

	/* set the current sensor as non-active */
	vrAssign6sensorActiveValue(((vr6sensor *)aux->sensor6_inputs[aux->active_sensor]), 0);

	/* search from the next possible sensor to the end of the list */
	do {
		aux->active_sensor++;
	} while (aux->sensor6_inputs[aux->active_sensor] == NULL && aux->active_sensor < MAX_6SENSORS);

	/* if none found (ie. we were basically already at the end of the list), then search from the beginning */
	if (aux->sensor6_inputs[aux->active_sensor] == NULL || aux->active_sensor >= MAX_6SENSORS) {
		for (aux->active_sensor = 0; aux->active_sensor < MAX_6SENSORS && aux->sensor6_inputs[aux->active_sensor] == NULL; aux->active_sensor++);
	}

	/* set the newly found sensor as active */
	vrAssign6sensorActiveValue(((vr6sensor *)aux->sensor6_inputs[aux->active_sensor]), 1);

	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: sensor6[%d] now active.\n", aux->active_sensor);
}

/* TODO: see if there is a way to call this as an N-switch */
/************************************************************/
static void _JoydevSensorSetCallback(vrInputDevice *devinfo, int value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	if (value == aux->active_sensor)
		return;

	if (value < 0 || value >= MAX_6SENSORS) {
		vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: set sensor (%d) -- out of range.\n", value);
	}

	if (aux->sensor6_inputs[value] == NULL) {
		vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: set sensor (%d) -- no such sensor available.\n", value);
		return;
	}

	vrAssign6sensorActiveValue(((vr6sensor *)aux->sensor6_inputs[aux->active_sensor]), 0);
	aux->active_sensor = value;
	vrAssign6sensorActiveValue(((vr6sensor *)aux->sensor6_inputs[aux->active_sensor]), 1);

	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: 6sensor[%d] now active.\n", aux->active_sensor);
}

/************************************************************/
static void _JoydevSensorSet0Callback(vrInputDevice *devinfo, int value)
{	_JoydevSensorSetCallback(devinfo, 0); }

/************************************************************/
static void _JoydevSensorSet1Callback(vrInputDevice *devinfo, int value)
{	_JoydevSensorSetCallback(devinfo, 1); }

/************************************************************/
static void _JoydevSensorSet2Callback(vrInputDevice *devinfo, int value)
{	_JoydevSensorSetCallback(devinfo, 2); }

/************************************************************/
static void _JoydevSensorSet3Callback(vrInputDevice *devinfo, int value)
{	_JoydevSensorSetCallback(devinfo, 3); }

/************************************************************/
static void _JoydevSensorSet4Callback(vrInputDevice *devinfo, int value)
{	_JoydevSensorSetCallback(devinfo, 4); }

/************************************************************/
static void _JoydevSensorSet5Callback(vrInputDevice *devinfo, int value)
{	_JoydevSensorSetCallback(devinfo, 5); }

/************************************************************/
static void _JoydevSensorSet6Callback(vrInputDevice *devinfo, int value)
{	_JoydevSensorSetCallback(devinfo, 6); }

/************************************************************/
static void _JoydevSensorSet7Callback(vrInputDevice *devinfo, int value)
{	_JoydevSensorSetCallback(devinfo, 7); }

/************************************************************/
static void _JoydevSensorSet8Callback(vrInputDevice *devinfo, int value)
{	_JoydevSensorSetCallback(devinfo, 8); }

/************************************************************/
static void _JoydevSensorSet9Callback(vrInputDevice *devinfo, int value)
{	_JoydevSensorSetCallback(devinfo, 9); }

/************************************************************/
static void _JoydevSensorResetCallback(vrInputDevice *devinfo, int value)
{
static	vrMatrix		idmat = { { 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0 } };
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;
	vr6sensor		*sensor;

	if (value == 0)
		return;

	sensor = &(devinfo->sensor6[aux->active_sensor]);
	vrAssign6sensorValue(sensor, &idmat, 0 /* , timestamp */);
	vrAssign6sensorActiveValue(sensor, -1);
	vrAssign6sensorErrorValue(sensor, 0);
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: reset 6sensor[%d].\n", aux->active_sensor);
}

/************************************************************/
static void _JoydevSensorResetAllCallback(vrInputDevice *devinfo, int value)
{
static	vrMatrix		idmat = { { 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0 } };
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;
	vr6sensor		*sensor;
	int			count;

	if (value == 0)
		return;

	for (count = 0; count < MAX_6SENSORS; count++) {
		if (aux->sensor6_inputs[count] != NULL) {
			sensor = &(devinfo->sensor6[count]);
			vrAssign6sensorValue(sensor, &idmat, 0 /* , timestamp */);
			vrAssign6sensorActiveValue(sensor, (count == aux->active_sensor));
			vrAssign6sensorErrorValue(sensor, 0);
			vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: reset 6sensor[%d].\n", count);
		}
	}
}

/************************************************************/
static void _JoydevTempValuatorOverrideCallback(vrInputDevice *devinfo, int value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	/* set the field to the current state of value (ie. 1 when depressed, 0 when released) */
	aux->sensor6_options.ignore_trans = value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: ignore_trans = %d.\n",
		aux->sensor6_options.ignore_trans);
}

/************************************************************/
static void _JoydevToggleValuatorOverrideCallback(vrInputDevice *devinfo, int value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.ignore_trans ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: ignore_trans = %d.\n",
		aux->sensor6_options.ignore_trans);
}

/************************************************************/
static void _JoydevTempValuatorOnlyCallback(vrInputDevice *devinfo, int value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	/* set the field to the current state of value (ie. 1 when depressed, 0 when released) */
	aux->sensor6_options.ignore_all = value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: ignore_all = %d.\n",
		aux->sensor6_options.ignore_all);
}

/************************************************************/
static void _JoydevToggleValuatorOnlyCallback(vrInputDevice *devinfo, int value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.ignore_all ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: ignore_all = %d.\n",
		aux->sensor6_options.ignore_all);
}

/************************************************************/
static void _JoydevToggleRelativeAxesCallback(vrInputDevice *devinfo, int value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.relative_axis ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: relative_axis = %d.\n",
		aux->sensor6_options.relative_axis);
}

/************************************************************/
/* TODO: this should probably also go through all the 6sensor's  */
/*   and move them to be within the allowed workspace when space */
/*   restriction is turned on.                                   */
static void _JoydevToggleRestrictSpaceCallback(vrInputDevice *devinfo, int value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.restrict_space ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: restrict_space = %d.\n",
		aux->sensor6_options.restrict_space);
}

/************************************************************/
static void _JoydevToggleReturnToZeroCallback(vrInputDevice *devinfo, int value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.return_to_zero ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: return_to_zero = %d.\n",
		aux->sensor6_options.return_to_zero);
}

/************************************************************/
static void _JoydevToggleSwapTransRotCallback(vrInputDevice *devinfo, int value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.swap_transrot ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: swap_transrot = %d.\n",
		aux->sensor6_options.swap_transrot);
}

/************************************************************/
static void _JoydevToggleSwapYZ(vrInputDevice *devinfo, int value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.swap_yz ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: swap_yz = %d.\n",
		aux->sensor6_options.swap_yz);
}

/************************************************************/
static void _JoydevSetTransXCallback(vrInputDevice *devinfo, float *value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_X] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: set trans_x = %f.\n", *value);
}

/************************************************************/
static void _JoydevSetTransYCallback(vrInputDevice *devinfo, float *value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_Y] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: set trans_y = %f.\n", *value);
}

/************************************************************/
static void _JoydevSetTransZCallback(vrInputDevice *devinfo, float *value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_Z] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: set trans_z = %f.\n", *value);
}

/************************************************************/
static void _JoydevSetAzimuthCallback(vrInputDevice *devinfo, float *value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_AZIM+3] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: set azimuth = %f.\n", *value);
}

/************************************************************/
static void _JoydevSetElevationCallback(vrInputDevice *devinfo, float *value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_ELEV+3] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: set elevation = %f.\n", *value);
}

/************************************************************/
static void _JoydevSetRollCallback(vrInputDevice *devinfo, float *value)
{
	_JoydevPrivateInfo	*aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_ROLL+3] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Joydev Control: set roll = %f.\n", *value);
}



	/*************************************************/
	/*   Callbacks for interfacing with the device.  */
	/*                                               */


/************************************************************/
static void _JoydevCreateFunction(vrInputDevice *devinfo)
{
	/*** List of possible inputs ***/
static	vrInputFunction	_JoydevInputs[] = {
				{ "button", VRINPUT_2WAY, _JoydevButtonInput },
				{ "axis", VRINPUT_VALUATOR, _JoydevValuatorInput },
				{ "valuator", VRINPUT_VALUATOR, _JoydevValuatorInput },
				{ "slider", VRINPUT_VALUATOR, _JoydevValuatorInput },
				{ "sim6", VRINPUT_6SENSOR, _Joydev6sensorInput },
				{ NULL, VRINPUT_UNKNOWN, NULL } };
	/*** List of control functions ***/
static	vrControlFunc	_JoydevControlList[] = {
				/* overall system controls */
				{ "system_pause_toggle", _JoydevSystemPauseToggleCallback },

				/* informational output controls */
				{ "print_context", _JoydevPrintContextStructCallback },
				{ "print_config", _JoydevPrintConfigStructCallback },
				{ "print_input", _JoydevPrintInputStructCallback },
				{ "print_struct", _JoydevPrintStructCallback },
				{ "print_help", _JoydevPrintHelpCallback },

				/* simulated 6-sensor selection controls */
				{ "sensor_next", _JoydevSensorNextCallback },
				{ "setsensor", _JoydevSensorSetCallback },	/* NOTE: this is non-boolean */
				{ "setsensor(0)", _JoydevSensorSet0Callback },
				{ "setsensor(1)", _JoydevSensorSet1Callback },
				{ "setsensor(2)", _JoydevSensorSet2Callback },
				{ "setsensor(3)", _JoydevSensorSet3Callback },
				{ "setsensor(4)", _JoydevSensorSet4Callback },
				{ "setsensor(5)", _JoydevSensorSet5Callback },
				{ "setsensor(6)", _JoydevSensorSet6Callback },
				{ "setsensor(7)", _JoydevSensorSet7Callback },
				{ "setsensor(8)", _JoydevSensorSet8Callback },
				{ "setsensor(9)", _JoydevSensorSet9Callback },
				{ "sensor_reset", _JoydevSensorResetCallback },
				{ "sensor_resetall", _JoydevSensorResetAllCallback },

				/* simulated 6-sensor manipulation controls */
				{ "temp_valuator", _JoydevTempValuatorOverrideCallback },
				{ "toggle_valuator", _JoydevToggleValuatorOverrideCallback },
				{ "temp_valuator_only", _JoydevTempValuatorOnlyCallback },
				{ "toggle_relative", _JoydevToggleRelativeAxesCallback },
				{ "toggle_space_limit", _JoydevToggleRestrictSpaceCallback },
				{ "toggle_return_to_zero", _JoydevToggleReturnToZeroCallback },
				{ "toggle_swap_transrot", _JoydevToggleSwapTransRotCallback },
				{ "toggle_swap_yz", _JoydevToggleSwapYZ },
				{ "set_transx", _JoydevSetTransXCallback },
				{ "set_transy", _JoydevSetTransYCallback },
				{ "set_transz", _JoydevSetTransZCallback },
				{ "set_azim", _JoydevSetAzimuthCallback },
				{ "set_elev", _JoydevSetElevationCallback },
				{ "set_roll", _JoydevSetRollCallback },

				/* other controls */
				/*   NONE   */

				/* end of the list */
				{ NULL, NULL } };

	_JoydevPrivateInfo	*aux = NULL;

	/******************************************/
	/* allocate and initialize auxillary data */
	devinfo->aux_data = (void *)vrShmemAlloc0(sizeof(_JoydevPrivateInfo));
	aux = (_JoydevPrivateInfo *)devinfo->aux_data;
	_JoydevInitializeStruct(aux, devinfo->type);

	/******************/
	/* handle options */
	aux->devfile = vrShmemStrDup(DEFAULT_DEVICE);	/* default, if no device file given */
	_JoydevParseArgs(aux, devinfo->args);

	/***************************************/
	/* create the inputs and self-controls */
	vrInputCreateDataContainers(devinfo, _JoydevInputs);
	vrInputCreateSelfControlContainers(devinfo, _JoydevInputs, _JoydevControlList);
	/* TODO: anything to do for NON self-controls?  Implement for Magellan first. */

	vrDbgPrintf("_JoydevCreateFunction(): Done creating Joydev inputs for '%s'\n", devinfo->name);
	devinfo->created = 1;

	return;
}


/************************************************************/
static void _JoydevOpenFunction(vrInputDevice *devinfo)
{
	_JoydevPrivateInfo	*aux = NULL;

	aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	/*******************/
	/* open the device */
#ifdef __linux /* { */
	aux->fd = open(aux->devfile, O_RDONLY);
	vrDbgPrintfN(INPUT_DBGLVL, "_JoydevOpenFunction(): just opened aux->fd = %d, devfile = '%s'\n", aux->fd, aux->devfile);
	if (aux->fd < 0) {
		vrErrPrintf("_JoydevOpenFunction(): Joydev device '%s' error: " RED_TEXT "couldn't open joydev device '%s'\n" NORM_TEXT,
			devinfo->name, aux->devfile);
		aux->open = 0;
		sprintf(aux->version, "- unconnected Joydev -");
	} else {
		if (_JoydevInitializeDevice(aux) < 0) {
			vrErrPrintf("_JoydevOpenFunction(): "
				RED_TEXT "Warning, unable to initialize Joydev '%s'.\n" NORM_TEXT, devinfo->name);
		}
		aux->open = 1;

		vrDbgPrintf("_JoydevOpenFunction(): Done opening Joydev input device '%s'.\n", devinfo->name);
		devinfo->ready = 1;
	}
#else /* } { ! linux */
	vrErrPrintf("_JoydevOpenFunction(): "
		RED_TEXT "Warning, unable to use Joydev devices on non-Linux systems. ('%s' inputs exist, but without hardware).\n" NORM_TEXT, devinfo->name);
	aux->open = 0;
	devinfo->ready = 0;
#endif /* } */

	return;
}


/************************************************************/
static void _JoydevCloseFunction(vrInputDevice *devinfo)
{
	_JoydevPrivateInfo	*aux = NULL;

	aux = (_JoydevPrivateInfo *)devinfo->aux_data;

	if (aux != NULL) {
		vrSerialClose(aux->fd);
		vrShmemFree(aux);	/* aka devinfo->aux_data */
	}

	return;
}


/************************************************************/
static void _JoydevResetFunction(vrInputDevice *devinfo)
{
#if 0 /* not yet used */
	_JoydevPrivateInfo	*aux = NULL;

	aux = (_JoydevPrivateInfo *)devinfo->aux_data;
#endif

	return;
}


/************************************************************/
static void _JoydevPollFunction(vrInputDevice *devinfo)
{
#if 0 /* not yet used */
	_JoydevPrivateInfo	*aux = NULL;

	aux = (_JoydevPrivateInfo *)devinfo->aux_data;
#endif

	if (devinfo->ready) {
		_JoydevGetData(devinfo);
	} else {
		/* TODO: try to open the device again */
	}

	return;
}



	/******************************/
	/*** FreeVR public routines ***/
	/******************************/


/******************************************************/
void vrJoydevInitInfo(vrInputDevice *devinfo)
{
	devinfo->version = (char *)vrShmemStrDup("-get from Joydev device-");
	devinfo->Create = vrCallbackCreateNamed("JoydevInput:Create-Def", _JoydevCreateFunction, 1, devinfo);
	devinfo->Open = vrCallbackCreateNamed("JoydevInput:Open-Def", _JoydevOpenFunction, 1, devinfo);
	devinfo->Close = vrCallbackCreateNamed("JoydevInput:Close-Def", _JoydevCloseFunction, 1, devinfo);
	devinfo->Reset = vrCallbackCreateNamed("JoydevInput:Reset-Def", _JoydevResetFunction, 1, devinfo);
	devinfo->PollData = vrCallbackCreateNamed("JoydevInput:PollData-Def", _JoydevPollFunction, 1, devinfo);
	devinfo->PrintAux = vrCallbackCreateNamed("JoydevInput:PrintAux-Def", _JoydevPrintStruct, 0);

	vrDbgPrintfN(INPUT_DBGLVL, "vrJoydevInitInfo: callbacks created.\n");
}


#endif /* } FREEVR */



	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/



#if defined(TEST_APP) /* { */

/******************************************************************/
/* Ugh, I hate globals, but I don't know a better way to get the  */
/*   aux value into the interrupt signal function exit_testapp(). */
static	int	done = 0;


/*******************************************************************/
void exit_testapp()
{
	done = 1;
}


/*******************************************************************/
/* A test program to communicate with a Joydev device and print the results. */
main(int argc, char *argv[])
{
	_JoydevPrivateInfo	*aux;
	int			count;

	done = 0;
	signal(SIGINT, exit_testapp);


	/******************************/
	/* setup the device structure */
	aux = (_JoydevPrivateInfo *)malloc(sizeof(_JoydevPrivateInfo));
	memset(aux, 0, sizeof(_JoydevPrivateInfo));
	_JoydevInitializeStruct(aux, "Joydevsubtype");


	/*********************************************************/
	/* set default parameters based on environment variables */
	aux->devfile = getenv("JOYDEV_DEVICE");
	if (aux->devfile == NULL)
		aux->devfile = DEFAULT_DEVICE;		/* default, if no file given */


	/*****************************************************/
	/* adjust parameters based on command-line-arguments */
	/* TODO: parse CLAs for non-default serial port, baud, etc. */


	/**************************************************/
	/* open the device file and initialize the device */
	aux->fd = open(aux->devfile, O_RDONLY);
	if (aux->fd < 0) {
		fprintf(stderr, RED_TEXT "couldn't open joydev device %s\n" NORM_TEXT, aux->devfile);
		aux->open = 0;
		sprintf(aux->version, "- unconnected Joydev -");
	} else {
		if (_JoydevInitializeDevice(aux) < 0) {
			vrErrPrintf("main: "
				RED_TEXT "Warning, unable to initialize Joydev.\n" NORM_TEXT);
		}
		aux->open = 1;
	}

	_JoydevPrintStruct(stdout, aux, verbose);

	/**********************/
	/* display the output */
	while(aux->open && !done) {
		if (_JoydevReadInput(aux) > 0) {
			printf("buttons: ");
			for (count = 0; count < aux->num_buttons; count++)
				printf("%d", aux->button[count]);
#if 0 /* text output of all the values */
			printf("  axes: ");
			for (count = 0; count < aux->axes; count++)
				printf("%5.1f", aux->axis[count]);
#else /* a little fancier text output */
			printf("  axes: |");
			for (count = 0; count < aux->num_axes; count++)
				printf("%5s|",
					(aux->axis[count] < -0.6 ? "-    " :
					 (aux->axis[count] < -0.3 ? " -   " :
					  (aux->axis[count] <  0.3 ? "  +  " :
					   (aux->axis[count] <  0.6 ? "   - " :
					    "    -")))));
#endif
			printf("    \r");
			fflush(stdout);
		}
	}

	printf("\n");

	/*****************/
	/* close up shop */
	if (aux != NULL) {
		close(aux->fd);
		free(aux);			/* aka devinfo->aux_data */
	}

#if 0
	_JoydevCloseDevice(aux);
	vrPrintf(BOLD_TEXT "\nJoydev device closed\n" NORM_TEXT);
#endif
}

#endif /* } TEST_APP */



	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/



#if defined(CAVE) /* { */


	/* ... CAVE stuff here if to also work with CAVElib */


#endif /* } CAVE */

